from ropchain.gadgets.asm import lea, xor, xchg, dec, orGadget, andGadget
from ropchain.gadgets import util, gadget
from ropchain import ropchain

'''
mov r1, r2
| lea r1, [r2]; ret
| lea r1, [r2+imm]; ret; (dec r1; ret)*
| xor r1, r1; ret; xor r1, r2; ret
| xor r1, r1; ret; add r1, r2; ret
| xor r1, r1; ret; or r1, r2; ret
| xchg r1, r2; ret mov r2, r1; ret; xchg r1, r2; ret
'''
def findWithoutXchg(r1, r2, gadgets, canUse):
    rop = gadget.find(gadgets, 'mov', r1, r2)
    if rop != None:
        rop = ropchain.ROPChain(rop)
        return rop
    rop = util.optMap(rop, fromLea, r1, r2, gadgets, canUse)
    rop = util.optMap(rop, fromLeaWithOffset, r1, r2, gadgets, canUse)
    rop = util.optMap(rop, fromXorXor, r1, r2, gadgets, canUse)
    rop = util.optMap(rop, fromXorOr, r1, r2, gadgets, canUse)
    return rop

def find(r1, r2, gadgets, canUse):
    rop = findWithoutXchg(r1, r2, gadgets, canUse)
    rop = util.optMap(rop, fromXchg, r1, r2, gadgets, canUse)
    return rop


'''
lea r1, [r2]
'''
def fromLea(r1, r2, gadgets, canUse):
    _lea = lea.find(r1, '[%s]' % r2, gadgets, canUse)
    if _lea != None:
        return ropchain.ROPChain(_lea)
    else:
        return None

'''
lea r1, [r2+n]
(dec r1) * n
'''
def fromLeaWithOffset(r1, r2, gadgets, canUse):
    #TODO replace find.ByRegex with wrapper function of lea
    _lea, _, imm = gadget.findByRegex(gadgets, 'lea', r1, r'\[%s\+0x[0-9a-fA-F]*]')
    _dec = dec.find(r1, gadgets, canUse)
    if _lea != None and _dec != None:
        return ropchain.ROPChain([_lea] +  [_dec] * imm)
    else:
        return None

'''
xor r1, r1;
xor r1, r2;
'''
def fromXorXor(r1, r2, gadgets, canUse):
    xorR1R1 = xor.find(r1, r1, gadgets, canUse)
    xorR1R2 = xor.find(r1, r2, gadgets, canUse)
    if xorR1R1 != None and xorR1R2 != None:
        return ropchain.ROPChain([xorR1R1, xorR1R2])
    else:
        return None

'''
xor r1, r1
or r1, r2
'''
def fromXorOr(r1, r2, gadgets, canUse):
    xorR1R1 = xor.find(r1, r1, gadgets, canUse)
    _or = orGadget.find(r1, r2, gadgets, canUse)
    if xorR1R1 != None and _or != None:
        return xorR1R1 + orGadget
    else:
        return None

'''
xchg r1, r2
mov r2, r1
xchg r1, r2
'''
def fromXchg(r1, r2, gadgets, canUse):
    _xchg = xchg.find(r1, r2, gadgets, canUse)
    _mov = findWithoutXchg(r2, r1, gadgets, canUse)
    if _xchg != None and _mov != None:
        return _xchg + _mov + _xchg
    else:
        return None



from ropchain import ropchain
from ropchain.gadgets import gadget, util
from ropchain.gadgets.asm import xor, add, inc, dec, orGadget, andGadget, movzx

def find(reg, gadgets, canUse):
    rop = fromXor(reg, gadgets, canUse)
    rop = util.optMap(rop, fromAddRegReg, reg, gadgets, canUse)
    rop = util.optMap(rop, fromOrAndDec, reg, gadgets, canUse)
    rop = util.optMap(rop, fromMovzx, reg, gadgets, canUse)
    return rop

def fromXor(reg, gadget, canUse):
    return xor.find(reg, reg, gadget, canUse)

def fromAddRegReg(reg, gadgets, canUse):
    addRegReg =  add.find(reg, reg, gadgets, canUse)
    if addRegReg != None:
        return addRegReg * 32
    else:
        return None

#move to make 0 routine ?
'''
or reg 1; ret; and reg 1; dec
'''
def fromOrAndDec(reg, gadgets, canUse):
    _or = orGadget.find(reg, '0x1', gadgets, canUse)
    _and = andGadget.find(reg, '0x1', gadgets, canUse)
    _dec = dec.find(reg, gadgets, canUse)
    if _or != None and _and != None and _dec != None:
        return _or + _and + _dec
    return None

'''
mov reg[:8], imm; ret; (inc reg; ret)*; movzx reg, reg[:8]; ret
'''
def fromMovzx(reg, gadgets, canUse):
    movLimm, imm, _ = gadget.findByRegex(gadgets, 'mov', '%s' % util.toL8bitReg(reg), r'0x[0-9a-fA-Z]+')
    _inc = inc.find(reg, gadgets, canUse)
    _movzx = movzx.find(reg, util.toL8bitReg(reg), gadgets, canUse)
    if movLimm != None and _inc != None and _movzx != None:
        return ropchain.ROPChain([movLimm] + _inc * (0x100 - imm) + _movzx)
    else:
        return None




from ropchain import ropchain
from ropchain.gadgets import util, gadget

def find(r1, r2, gadgets, canUse):
    rop = util.optROPChain(gadget.find(gadgets, 'xchg', r1, r2))
    rop = rop if rop != None else util.optROPChain(gadget.find(gadgets, 'xchg', r2, r1))
    return rop



from ropchain.gadgets import util
from ropchain.gadgets.asm import add, shl, mov

def find(reg, gadgets, canUse):
    rop = add.find(reg, reg, gadgets, canUse)
    if rop != None:
        return rop

    rop = util.optMap(rop, fromOtherReg, reg, gadgets, canUse)
    rop = util.optMap(rop, fromShiftL, reg, gadgets, canUse)
    return None

def fromOtherReg(reg, gadgets, canUse):
    for r in canUse:
        _mov = mov.find(r, reg, gadgets, canUse - set(r))
        _add = add.find(reg, r, gadgets, canUse)
        if _mov != None and _add != None:
            return _mov + _add
    return None

def fromShiftL(reg, gadgets, canUse):
    return shl.find(reg, '0x1', gadgets, canUse)



from ropchain.gadgets import gadget, util
from ropchain import ropchain

#TODO seprate add reg, reg; add reg, imm
def find(op1, op2, gadgets, canUse):
    rop = util.optROPChain(gadget.find(gadgets, 'add', op1, op2))
    return rop

# def fromInc(op1, gadgets, canUse):
#     _inc = inc.find(op1, gadgets, canUse)



from solver import solve
from gadgets import gadget
from pwn import asm
import emulator
import unittest

class TestROPChain(unittest.TestCase):
    def do(self, dests, gadgetsDict):
        lib = buildFromGadgets(gadgetsDict)
        gadgets = gadget.fromDict(gadgetsDict)
        payload = solve(dests, gadgets, emulator.LIB_BASE).payload()
        testResult = emulator.execROPChain(payload, lib)
        self.assertTrue(isCorrect(dests, testResult))

    def testPop(self):
        dests = {'eax': 0x0, 'ebx': 0x602030, 'ecx': 0x1000}
        gadgetsDict = {
                0x10000: 'pop eax; ret',
                0x20000: 'pop ebx; ret',
                0x30000: 'pop ecx; ret'
                }
        self.do(dests, gadgetsDict)

    def testIncAdd(self):
        dests = {'eax': 0x12345678}
        gadgets = {
                0x10000: 'xor eax, eax; ret',
                0x20000: 'inc eax; ret',
                0x30000: 'add eax, eax; ret'
            }
        self.do(dests, gadgets)

    def testPopMov(self):
        dests = {'eax': 0x12345678, 'esi': 0x41414141}
        gadgets = {
                0x1000: 'pop esi; ret',
                0x2000: 'mov eax, esi; ret'
            }
        self.do(dests, gadgets)

    def testPopMov6Regs(self):
        dests = {
            'eax': 0x12345678,
            'ebx': 0x87654321,
            'ecx': 0x22222222,
            'edx': 0x7fffffff,
            'esi': 0x41414141,
            'edi': 0x42424242
        }
        gadgets = {
            0x1000: 'pop esi; ret',
            0x2000: 'mov ebx, esi; ret',
            0x3000: 'mov eax, ecx; ret',
            0x4000: 'mov ecx, ebx; ret',
            0x5000: 'mov edx, esi; ret',
            0x6000: 'mov edi, eax; ret',
        }
        self.do(dests, gadgets)

    def testMovWithoutPop(self):
        dests = {'eax':  0x12345678, 'esi':  0x41414141}
        gadgets = {
            0x1000: 'xor esi, esi; ret',
            0x2000: 'inc esi; ret',
            0x3000: 'add esi, esi; ret',
            0x4000: 'mov eax, esi; ret',
        }
        self.do(dests, gadgets)

    def testMovWithoutPop6Regs(self):
        dests = {
            'eax':  0x12345678,
            'ebx':  0x22222222,
            'ecx':  0x33333333,
            'edx':  0x87654321,
            'esi':  0x41414141,
            'edi':  0x42424242,
        }
        gadgets = {
            0x1000: 'xor esi, esi; ret',
            0x2000: 'inc esi; ret',
            0x3000: 'add esi, esi; ret',
            0x4000: 'mov eax, esi; ret',
            0x5000: 'mov ebx, eax; ret',
            0x6000: 'mov edx, ecx; ret',
            0x7000: 'mov edi, edx; ret',
            0x8000: 'mov ecx, ebx; ret',
        }
        self.do(dests, gadgets)

    def testPopXor(self):
        dests = {'eax': 0x12345678, 'esi': 0x41414141}
        gadgets = {
                0x1000: 'pop esi; ret',
                0x2000: 'xor eax, eax; ret',
                0x3000: 'xor eax, esi; ret'
            }
        self.do(dests, gadgets)

def buildFromGadgets(gadgets):
    addrs = gadgets.keys()
    insns = list(map(asm, gadgets.values()))
    lib = '\x00' * (max(addrs) + len(max(insns, key=len)))
    for addr, insn in zip(addrs, insns):
        lib = lib[:addr] + insn + lib[addr+len(insn):]
    return lib

def isCorrect(expected, result):
    for reg  in expected:
        if result[reg] != expected[reg]:
            return False
    return True

if __name__ == '__main__':
    unittest.main()



from ropchain import ropchain
def optMap(x, y, *args):
    if x != None:
        return x
    else:
        return y(*args)

optROPChain = lambda x: ropchain.ROPChain(x) if x != None else None

def toH8bitReg(reg):
    if reg == 'eax':
        return 'ah'
    elif reg == 'ebx':
        return 'bh'
    elif reg == 'ecx':
        return 'ch'
    elif reg == 'edx':
        return 'dh'

def toL8bitReg(reg):
    if reg == 'eax':
        return 'al'
    elif reg == 'ebx':
        return 'bl'
    elif reg == 'ecx':
        return 'cl'
    elif reg == 'edx':
        return 'dl'



from ropchain import ropchain
from ropchain.gadgets.asm import mov, xor, inc, xchg, add, double
from ropchain.gadgets import util, gadget, toZero

'''
alt pop reg
| mov reg, imm; ret; (add reg, reg; ret;)*; ([inc reg], add reg, reg)*; ret
| mov reg, other; ret; (add reg, reg; ret;)*; ([inc reg], add reg, reg)*; ret
| xor reg, reg; ret; ([inc reg], add reg, reg)*; ret
| pop other; ret; mov reg, other; ret
| pop other; ret; xchg reg, other; ret
'''

def find(reg, dest, gadgets, canUse):
    rop = gadget.find(gadgets, 'pop', reg)
    if rop != None:
        rop =  ropchain.ROPChain(rop)
        rop.appendValue(dest)
        return rop

    rop = util.optMap(rop, fromIncAdd, dest, reg, gadgets, canUse)
    rop = util.optMap(rop, fromOtherReg, dest, reg, gadgets, canUse)

    return rop

'''
xor reg, reg; ret; ([inc reg], add reg, reg)*; ret
'''
def fromIncAdd(dest, reg, gadgets, canUse):
    zero = toZero.find(reg, gadgets, canUse)
    _inc = inc.find(reg, gadgets, canUse)
    _double = double.find(reg, gadgets, canUse)

    if zero != None and _inc != None and _double != None:
        ret = zero
        bits = bin(dest)[2:]
        bits = '0' * (32 - len(bits)) + bits
        for i in range(31):
            if bits[i] == '1':
                ret += _inc
            ret += _double
        if bits[31] == '1':
            ret += _inc

        return ret

    return None

'''
pop other; ret; mov reg, other; ret
pop other; ret; xchg reg, other; ret
'''
def fromOtherReg(dest, reg, gadgets, canUse):
    for r in canUse:
        _pop = find(r, dest, gadgets, canUse - set([r]))
        _mov = mov.find(reg, r, gadgets, canUse - set([reg, r]))

        if _pop != None and _mov != None:
            return _pop + _mov

        _xchg = xchg.find(reg, r, gadgets, canUse - set(r))
        if _pop != None and _xchg != None:
            return _pop + _xchg

    return None



from ropchain import frontend
from subprocess import call
import re
import os
import md5

def find(gadgets, mnem, op1=None, op2=None, op3=None):
    for gadget in gadgets:
        ops = list(filter(lambda x: x != None, [op1, op2, op3]))
        if gadget.eq(mnem, ops):
            return gadget
    return None

def findByRegex(gadgets, mnem, op1=None, op2=None):
    for gadget in gadgets:
        for i, _mnem in enumerate(gadget.mnems):
            if not re.match(mnem, _mnem):
                continue
            ops = gadget.ops[i]
            if op1 != None and not re.match(op1, ops[0]):
                continue
            if op2 != None and not re.match(op2, ops[1]):
                continue
            return mnem, ops[0], ops[1]
    return None, None, None

class Gadget:
    def __init__(self, gadgets, addr=0):
        gadgets = list(filter(lambda x: len(x) > 0, gadgets))
        self.addr = addr
        self.mnems = []
        self.ops = []
        for gadget in gadgets:
            ls = gadget.split()
            ls = list(map(lambda x: x.strip(), ls))
            mnem = ls[0]
            ops = list(map(lambda x: x.strip(), ' '.join(ls[1:]).split(",")))
            self.mnems.append(mnem)
            self.ops.append(ops)

    def puts(self, base=0):
        print(hex(self.addr + base) + " " + '; '.join(map(lambda x: "%s %s" % (x[0], ', '.join(x[1])), zip(self.mnems, self.ops))))

    def eq(self, _mnem, _ops):
        return any([mnem == _mnem and ops == _ops for mnem, ops in zip(self.mnems, self.ops)])


def readGadgets(filePath):
    return open(filePath).readlines()

def fromDict(gadgets):
    return [Gadget(insn.split(';'), addr) for addr, insn in gadgets.iteritems()]

#HACKME
def load(filePath):
    #check gadgetfile
    md5sum = ''.join(map(lambda x: "%02x" % ord(x), md5.new(open(filePath).read()).digest()))
    hashFullPath = "./.cache/ropchain/%s" % md5sum
    if not os.path.exists(hashFullPath):
        fullPathFrontEnd = os.path.dirname(os.path.abspath(frontend.__file__))
        call(["bash", "%s/ropgadget.sh" % fullPathFrontEnd, filePath])
    lines = readGadgets(hashFullPath)
    return parseGadget(lines)


def parseGadget(txtGadget):
    addrs = list(map(lambda x: int(x.replace(":", "").split()[0], 16), txtGadget))
    gadgets = list(map(lambda x: ' '.join(x.split()[1:]), txtGadget))
    gadgets = list(map(lambda x: list(map(lambda y: y.strip(),x.split(';'))), gadgets))
    gadgets, addrs = zip(*list(filter(lambda xs: any([x == 'ret' for x in xs[0]]), zip(gadgets, addrs))))
    return list(map(lambda x: Gadget(x[0], x[1]), zip(gadgets, addrs)))




import ropchain
from gadgets import gadget, setVal
import itertools
import copy

def solve(dests, gadgets, base=0):
    regs = set(dests.keys())
    ropChains = {reg: calcByOneReg(dests[reg], reg, gadgets) for reg in regs}
    ropChains = dict((k, v) for k, v in ropChains.iteritems() if v)
    solvable = set(ropChains.keys())

    remains = regs - solvable
    ans = None
    for rs in itertools.permutations(remains):
        tmpAns = ropchain.ROPChain([])
        canUse = copy.deepcopy(regs)
        for reg in rs:
            canUse.remove(reg)
            tmp = calcWithOtherRegs(dests[reg], reg, gadgets, canUse)
            if tmp == None:
                tmpAns = None
                break
            tmpAns += tmp

        if ans == None:
            ans = tmpAns
        elif tmpAns == None:
            ans = ans
        elif len(ans.payload()) > len(ans.payload()):
            ans = tmpAns

    r = sum([ropChains[reg] for reg in ropChains], ans)
    r.setBase(base)
    return r

def calcByOneReg(dest, reg, gadgets):
    return setVal.find(reg, dest, gadgets, set())

def calcWithOtherRegs(dest, reg, gadgets, canUse):
    return setVal.find(reg, dest, gadgets, canUse)



from gadgets.gadget import Gadget
from struct import pack

class ROPChain:
    def __init__(self, gadgets, base=0):
        self.gadgets = []
        self.base = base

        if not isinstance(gadgets, list):
            gadgets = [gadgets]

        for gadget in gadgets:
            if gadget == None:
                continue

            if isinstance(gadget, ROPChain):
                self.chain(gadget)
            else:
                self.gadgets.append(gadget)

    def appendGadget(self, gadget):
        self.gadgets.append(gadget)

    def appendValue(self, value):
        gadget = Gadget([], value)
        self.gadgets.append(gadget)

    def dump(self):
        for gadget in self.gadgets:
            if self.isGadget(gadget):
                gadget.puts(self.base)
            else:
                gadget.puts()

    def setBase(self, base):
        self.base = base

    def payload(self):
        payload = ''
        for gadget in self.gadgets:
            if self.isGadget(gadget):
                payload += p32(gadget.addr + self.base)
            else:
                payload += p32(gadget.addr)
        return payload

    def chain(self, ropChain):
        self.gadgets += ropChain.gadgets

    def isGadget(self, gadget):
        return len(gadget.mnems) > 0

    def __iadd__(self, ropChain):
        self.chain(ropChain)
        return self

    def __add__(self, ropChain):
        self.gadgets += ropChain.gadgets
        return self

    def __radd__(self, ropChain):
        return self.__add__(ropChain)

    def __mul__(self, times):
        self.gadgets *= times
        return self
def p32(n):
    return pack("<I", n)



from ropchain import ropchain
from ropchain.gadgets import gadget

def find(reg, gadgets, canUse):
    dec = gadget.find(gadgets, 'dec', reg)
    if dec != None:
        return ropchain.ROPChain(dec)
    else:
        return None




from ropchain.gadgets import util, gadget
from ropchain.gadgets.asm import inc
from ropchain import ropchain

def find(r1, r2, gadgets, canUse):
    xor = util.optROPChain(gadget.find(gadgets, 'xor', r1, r2))

    return xor



from setuptools import setup, find_packages
from setuptools.command.install import install
from subprocess import call
import os
import sys

class CustomInstallCommand(install):
    def run(self):
        currentModule = sys.modules[__name__]
        dirName = os.path.dirname(os.path.abspath(currentModule.__file__))
        scriptPath = "%s/require.sh" % dirName
        call(['bash', scriptPath])
        install.run(self)
 
setup(
        name             = 'ropchain',
        version          = '0.1.3',
        description      = 'ROPChain generator',
        license          = 'GPL3.0',
        author           = 'kriw',
        author_email     = 'kotarou777775@gmail.com',
        url              = 'https://github.com/kriw/ropchain',
        keywords         = '',
        packages         = find_packages(),
        cmdclass         = {'install': CustomInstallCommand},
        include_package_data = True
        )



from ropchain.gadgets import util, gadget
from ropchain.gadgets.asm import neg, dec, add, lea
from ropchain import ropchain

def find(reg, gadgets, canUse):
    rop = util.optROPChain(gadget.find(gadgets, 'inc', reg))
    rop = util.optMap(rop, fromNegDec, reg, gadgets, canUse)
    rop = util.optMap(rop, fromLeaPlusOne, reg, gadgets, canUse)
    rop = util.optMap(rop, fromAddOne, reg, gadgets, canUse)
    return rop

def fromNegDec(reg, gadgets, canUse):
    _neg = neg.find(reg, gadgets, canUse)
    _dec = dec.find(reg, gadgets, canUse)
    if _neg != None and _dec != None:
        return ropchain.ROPChain([_neg, _dec, _dec, _neg])
    else:
        return None

def fromAddOne(reg, gadgets, canUse):
    addOne = add.find(reg, '0x1', gadgets, canUse)
    if addOne != None:
        return ropchain.ROPChain(addOne)
    else:
        return None


def fromLeaPlusOne(reg, gadgets, canUse):
    for r in canUse:
        _lea = lea.find(reg, '[%s+0x1]' % r, gadgets, canUse)
        # #FIXME importing module 'mov' make import cycle
        # _mov = mov.find(r, reg, gadgets, canUse - set([r]))
        _mov = util.optROPChain(gadget.find(gadgets, 'mov', r, reg))
        if _lea != None and _mov != None:
            return ropchain.ROPChain([_lea, _mov])
    return None



from ropchain import ropchain
from ropchain.gadgets import gadget 

def find(reg, gadgets, canUse):
    neg = gadget.find(gadgets, 'neg', reg)
    if neg != None:
        return ropchain.ROPChain(neg)
    else:
        return None




from unicorn import *
from unicorn.x86_const import *
from pwn import asm
import solver
from struct import unpack
import sys
# memory address where emulation starts
ADDRESS = 0x1000000
STACK_BASE = 0xf7000000
LIB_BASE = 0x7f000000
STACK_SIZE = 200 * 1024 * 1024
DEST = 0x0
get_eax = lambda :mu.reg_read(UC_X86_REG_EAX)
get_ebx = lambda :mu.reg_read(UC_X86_REG_EBX)
get_ecx = lambda :mu.reg_read(UC_X86_REG_ECX)
get_edx = lambda :mu.reg_read(UC_X86_REG_EDX)

def align(n):
    ret = 0x0;
    while ret < n:
        ret += 0x1000
    return ret

def init(payload, lib):
    LIB_SIZE = align(len(lib))

    # initialize unicorn emulator
    mu = Uc(UC_ARCH_X86, UC_MODE_32)
    mu.mem_map(ADDRESS, 0x1000)
    mu.mem_map(LIB_BASE, LIB_SIZE)
    mu.mem_map(STACK_BASE, STACK_SIZE)
    mu.mem_map(DEST, 0x1000)

    mu.reg_write(UC_X86_REG_ESP, STACK_BASE + STACK_SIZE / 2)
    mu.mem_write(ADDRESS, asm('ret'))
    mu.mem_write(LIB_BASE, lib)
    mu.mem_write(mu.reg_read(UC_X86_REG_ESP), payload)
    return mu

def execROPChain(payload, lib, output=False):
    mu = init(payload, lib)
    # emulate code in infinite time & unlimited instructions
    mu.emu_start(ADDRESS, DEST)

    r_eax = mu.reg_read(UC_X86_REG_EAX)
    r_ebx = mu.reg_read(UC_X86_REG_EBX)
    r_ecx = mu.reg_read(UC_X86_REG_ECX)
    r_edx = mu.reg_read(UC_X86_REG_EDX)
    r_ebp = mu.reg_read(UC_X86_REG_EBP)
    r_esp = mu.reg_read(UC_X86_REG_ESP)
    r_edi = mu.reg_read(UC_X86_REG_EDI)
    r_esi = mu.reg_read(UC_X86_REG_ESI)
    r_eip = mu.reg_read(UC_X86_REG_EIP)
    r_eflags = mu.reg_read(UC_X86_REG_EFLAGS)

    if output:
        print("=========HALT==========")
        print(">>> EAX = 0x%x" % r_eax)
        print(">>> EBX = 0x%x" % r_ebx)
        print(">>> ECX = 0x%x" % r_ecx)
        print(">>> EDX = 0x%x" % r_edx)
        print(">>> EBP = 0x%x" % r_ebp)
        print(">>> ESP = 0x%x" % r_esp)
        print(">>> EDI = 0x%x" % r_edi)
        print(">>> ESI = 0x%x" % r_esi)
        print(">>> EIP = 0x%x" % r_eip)
        print(">>> EFLAGS = 0x%x" % r_eflags)

    return {
        'eax': r_eax,
        'ebx': r_ebx,
        'ecx': r_ecx,
        'edx': r_edx,
        'ebp': r_ebp,
        'esp': r_esp,
        'edi': r_edi,
        'esi': r_esi,
        'eflags': r_eflags,
    }
